package unicrawler.crawler;

import org.apache.log4j.Logger;

import unicrawler.CrawlerEnvironment;
import unicrawler.types.CrawlerURL;
import unicrawler.types.Page;

public class CrawlerWorker implements Runnable {

	private Logger logger = Logger.getLogger(CrawlerWorker.class.getName());
	
	private CrawlerWorkerCounter mWorkerCounter;
	private CrawlerEnvironment mEnvironment;
	
	private CrawlerURL mUrl;
	
	public CrawlerWorker(CrawlerEnvironment _env, CrawlerURL _url, CrawlerWorkerCounter _counter) {
		
		this.mEnvironment = _env;
		this.mWorkerCounter = _counter;
		this.mUrl = _url;
		
		this.mWorkerCounter.addCrawlerCount();
	}
	
	@Override
	public void run() {
		// Lets start doing this crawler buisness...
		try {
			// Check if we're only crawling the seed URL domain
			if (this.mEnvironment.getConfig().onlyCrawlSeedDomain() && ! this.mUrl.getDomain().contains(this.mEnvironment.getSeedUrls().getDomain())) {
				
				// Don't crawl this page because we're only crawling the seed URL
				if (this.mEnvironment.getConfig().showSeedOnlyWarn()) {
					logger.warn("Not crawling " + this.mUrl.getDomain() + " - it doesnt match seed domain");
				}

				return;
			}
			
			Page crawledPage = this.mEnvironment.getDownloader().downloadPage(this.mUrl.getAsObject());
						
			if (crawledPage == null) {
				// Page wasn't downloaded
				if (this.mEnvironment.getConfig().showSkipped()) {
					logger.error(mUrl.getLink() + " was not downloaded.");
				}
				
				this.mEnvironment.incrementSkippedCount();
								
				return;
			}
			
			if (crawledPage.getPageResponseCode() == 200) {
				// Page was downloaded okay!
				this.mEnvironment.getFrontier().addSuccessfulCrawl(this.mUrl, crawledPage);

				// Parse the page links
				crawledPage.parsePageLinks(this.mUrl.getLink());
				
				for (String link : crawledPage.getPageLinks()) {	
					// Have we already seen this URL?
					if (this.mEnvironment.getFrontier().hasSeenUrl(link)) continue;
					
					// If we've reached the max depth - break out!
					if (this.mEnvironment.getConfig().getMaxCrawlDepth() > 0 && (this.mUrl.getDepth() + 1) > this.mEnvironment.getConfig().getMaxCrawlDepth()) {
						if (this.mEnvironment.getConfig().showMaxDepthWarn()) {
							logger.warn("Max depth reached for \"" + link + "\"");
						}
						
						return;
					}
										
					CrawlerURL url = new CrawlerURL(link, (this.mUrl.getDepth() + 1));
					
					// Is this URL within our Disallowed hosts by robots.txt?
					if ( this.mEnvironment.getConfig().useRobotManager() && ! this.mEnvironment.getRobotsManager().determineConstraint(url.getAsObject()) ) {
						// This URL isn't allowed!		
						System.out.println("robots.txt disallowed " + url.getLink());
						return;
					}

					// Schedule the url by adding to the frontier
					//System.out.println("queueing " + url.getLink());
					this.mEnvironment.getFrontier().enqueue(url);
				}
			}
			else if (crawledPage.getPageResponseCode() >= 300 && crawledPage.getPageResponseCode() < 400) {
				// Must have been some kind of redirect, re-schedule to frontier				
				CrawlerURL redirect = crawledPage.getRedirectUrl(this.mUrl.getDepth());
				if (redirect != null) {
					// Make sure the header contained our redirect...
					this.mEnvironment.getFrontier().enqueue(redirect);
				
					this.mEnvironment.incrementRedirectedCount();
				}
			}
			else {
				// Unknown error...
				if (this.mEnvironment.getConfig().showResponseErrors()) {
					logger.error(mUrl.getLink() + " obtained HTTP errors. Response code [" + crawledPage.getPageResponseCode() + "]");
				}
				this.mEnvironment.incrementResponseErrorCount();
			}

		} catch (Exception e) {
			logger.error(e);
			e.printStackTrace();
			
		}
		finally {
			// This worker has finished executing!
			this.mWorkerCounter.minusCrawlerCount();
		}
	}
}
