package unicrawler.downloader;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.SocketTimeoutException;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.zip.GZIPInputStream;

import org.apache.commons.io.IOUtils;
import org.apache.log4j.Logger;

import unicrawler.CrawlerEnvironment;
import unicrawler.config.CrawlerConfig;
import unicrawler.types.Page;

public class WebpageDownloader implements IDownloader {

	private Logger logger = Logger.getLogger(WebpageDownloader.class.getName());
	private CrawlerEnvironment mEnvironment;
	private CrawlerConfig mConfig;
	
	/* Downloader properties */
	private String userAgent;
	private int maxRetries;
	private boolean followRedirects;
	private String[] contentTypes = { };
		
	public WebpageDownloader(CrawlerEnvironment _env) {
		this.mEnvironment = _env;
		this.mConfig = mEnvironment.getConfig();
		
		this.userAgent = mConfig.getDownloaderUseragent();
		this.followRedirects = mConfig.doesDownloaderFollowRedirects();
		this.contentTypes = mConfig.getDownloaderAllowedContentTypes();
		this.maxRetries = mConfig.getDownloaderRetries();	
	}
	
	private DownloaderRequest createDownloadRequest(URL url) {
		DownloaderRequest request = new DownloaderRequest(url);
		
		// Add the standard headers as grabbed by Google Chrome
		request.addHeader("Accept-Language", "en-US,en;q=0.8");
		request.addHeader("Cache-Control", "no-cache");	// We don't want to cache pages!
		request.addHeader("Pragma", "no-cache");
		request.addHeader("Accept-Encoding", "gzip,deflate,sdch");
		request.addHeader("User-Agent", this.userAgent);
		request.addHeader("Connection", "close");	// no support for keep-alive atm
		
		return request;
	}
	
	private	 HttpURLConnection createHttpConnection(DownloaderRequest request) throws IOException {
		URL url = request.getRequestURL();
		HttpURLConnection http = (HttpURLConnection)url.openConnection();
		
		http.setInstanceFollowRedirects(this.followRedirects);
		http.setConnectTimeout(30000);	//30000ms
		http.setReadTimeout(30000);	//30000ms
		http.setRequestMethod("GET");
		
		// Add all our requestheaders from our downloader request object
		Iterator<Entry<String, String>> mapIterator = request.getRequestHeaders().entrySet().iterator();
		while (mapIterator.hasNext()) {
			Map.Entry<String, String> pair = (Map.Entry<String, String>)mapIterator.next();
			http.setRequestProperty(pair.getKey(), pair.getValue());
		}
		
		return http;
	}
	
	public Page downloadPage(URL _url) throws IOException {
		Page page = null;
		DownloaderRequest request = this.createDownloadRequest(_url);

		HttpURLConnection http = null;
		long startTime = 0;
		
		// For handling of proper response from server, obj Page will hold this data
		Map<String, String> serverHeaderResponses = new HashMap<String, String>();
		int respCode = 0;
		
		byte[] content = null;
		String contentEncoding = null;
		String contentCharset = null;
		
		for (int i = 0; i < this.maxRetries; i++) {
			try {
				startTime = System.currentTimeMillis();
				
				http = this.createHttpConnection(request);
				http.connect();
				
				// Time to gather connection info, make sure we have no errors
				respCode = http.getResponseCode();
								
				// Iterate through our server resposne headers to determine charset encoding etc
				for (String key : http.getHeaderFields().keySet()) {
					String value = http.getHeaderField(key);					
					serverHeaderResponses.put(key, value);
				}
				
				// Get content with appropriate content encoding (could be compressed)
				contentEncoding = http.getContentEncoding();

				InputStream is = http.getInputStream();
				// TODO: clean this, it's very ugly
				try {
					// Do we have compressed data? usually the browser uncompresses this for us
					if (contentEncoding != null && contentEncoding.equals("gzip")) {
						content = IOUtils.toByteArray(new GZIPInputStream(is));
					}
					else {
						content = IOUtils.toByteArray(is);
					}
				} finally {
					if (is != null) {
						try {
							is.close();
						}
						catch (IOException ex) {
							logger.warn("Unable to close http connection input stream!", ex);
						}
					}
				}
				
				// Check to make sure we have some actual content before procceeding
				if (content == null) {
					// We have no content... 
					logger.warn("There was no content for page " + request.getRequestURLString());
					
					if (i == (this.maxRetries - 1)) {
						logger.warn("Maximum retries was reached for " + request.getRequestURLString());
						return null;
					}
					else {
						continue;
					}
				}
				
				// Now to get the content character encoding...
				String header = http.getContentType();
				
				if (header != null) {
					String[] slugs = header.split(";");
					for (String slug : slugs) {
						slug = slug.trim();
						if (slug.toLowerCase().startsWith("charset=")) {
							contentCharset = slug.substring("charset=".length());
						}
					}
					
					// Check our content type is contained within the allowed types
					for (String type : this.contentTypes) {
						if ( ! http.getContentType().contains(type)) { 
							// As far as the crawlers concerned, we didn't download this.
							return null;
						}
					}
				}
								
				// TODO: get charset from MIME-TYPE if not found in content-type header
				if (contentCharset == null) {
					// Set to UTF-8
					contentCharset = "UTF-8";
				}
			}
			catch (SocketTimeoutException ex) {
				respCode = HttpURLConnection.HTTP_CLIENT_TIMEOUT;
			}
			catch (FileNotFoundException ex) {
				respCode = HttpURLConnection.HTTP_NOT_FOUND;
			}
			catch (IOException ex) {
				respCode = HttpURLConnection.HTTP_UNAVAILABLE;
			}
			finally {
				// Maybe check the Http URL Connection errorStream.. TODO?
				http.disconnect();
			}
			
			long respTime = System.currentTimeMillis() - startTime;
						
			page = new Page(mConfig, request.getRequestURL(), content, respCode, respTime, contentCharset, serverHeaderResponses);

			return page;
		}
		
		// this is almost certainly likely to be null - meaning page was skipped
		return page;
	}
}
